from .faker_generator import FakerGenerator
from .named_entity_recognizer import frequency, NamedEntityRecognizer, split_name


__all__ = ["FakerGenerator", "NamedEntityRecognizer", "split_name", "frequency"]



import re
from typing import Any, Dict, List, Union, Optional

from faker import Faker

import numpy as np
import pandas as pd
from simple_colors import green, red


class FakerGenerator:
    """
    A class used to generate faker objects in a dataframe

    Attributes
    -------
    dataset : pd.DataFrame
        A pandas dataframe
    dict_global_entities : Dict
        A dictionary whose keys have the same names of the dataframe columns and values
        are dictionaries in which the entity associated to the column and its confidence
        score are reported.
    faker : Any
        A generator to obtain synthetisized objects
    columns_with_assigned_entities : List
        A list of columns with an assigned entity
    columns_not_synthesized : List
        A list of those columns which are not synthesized by faker
    list_faker : List
        A list of those columns which are synthesized by faker
    lang : str
        A string to set italian as language of generation. Default: en
    generation_mark : Any
        A mark for the generation (usually *)

    Returns
    -------
    _type_
        _description_
    """

    dataset: pd.DataFrame
    dict_global_entities: Dict
    faker: Any
    columns_with_assigned_entity: List
    columns_not_synthesized: List
    list_faker: List
    lang: str
    generation_mark: str

    def __init__(
        self,
        df_input: Union[str, pd.DataFrame],
        dict_global_entities: Dict,
        lang: str = "en",
        generation_mark: Optional[str] = None,
    ) -> "FakerGenerator":
        """
        Create a FakerGenerator instance

        Parameters
        ----------
        df_input : Union[str, pd.DataFrame]
            A pandas dataframe or a path to a csv file.
        dict_global_entities : Dict
            A dictionary whose keys have the same names of the dataframe columns and
            values are dictionaries in which the entity associated to the column and
            its confidence score are reported.
        lang : str, optional
            Input language by default "en". Set this parameter to "it" to return better performance on Italian data.

        Returns
        -------
        FakerGenerator
            A Fakergenerator instance.
        """

        if not isinstance(df_input, pd.DataFrame):
            df_input = pd.read_csv(df_input)

        self.dataset = df_input
        self.dict_global_entities = dict_global_entities
        self.lang = lang
        if self.lang == "it":
            self.faker = Faker(["it_IT"])
        else:
            self.faker = Faker()
        self.columns_with_assigned_entity = []
        self.columns_not_synthesized = []
        self.list_faker = []
        self.generation_mark = generation_mark

    def get_columns_with_assigned_entity(self) -> None:
        """
        Get a list containing those columns with an assigned entity and confidence
        score > 0.3.

        """

        if len(self.dict_global_entities) > 0:
            columns_with_assigned_entity = [
                [i, self.dict_global_entities[i]["entity"]]
                for i in self.dict_global_entities
                if self.dict_global_entities[i] is not None
                and self.dict_global_entities[i]["confidence_score"] > 0.3
            ]
            self.columns_not_synthesized = [
                [i, self.dict_global_entities[i]["entity"]]
                for i in self.dict_global_entities
                if self.dict_global_entities[i] is not None
                and self.dict_global_entities[i]["confidence_score"] <= 0.3
                and not re.match(".*?last.*?name.*?", i.lower())
            ]

        if len(columns_with_assigned_entity) > 0:
            self.columns_with_assigned_entity = columns_with_assigned_entity
        else:
            return print("Impossible to generate Faker data: no assigned entities.")

    def get_address(self) -> None:
        """
        Synthesize address columns in a pandas dataframe

        """

        addresses = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "ADDRESS"
            or ("indirizzo" in i[0].lower())
            or (
                (i[1] == "LOCATION")
                and (("address" in i[0].lower()) or (("indirizzo" in i[0].lower())))
            )
        ]

        for i in addresses:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.street_address()
                        if row == self.generation_mark
                        else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.street_address() if not pd.isnull(row) else np.NaN
                    )
                )

                self.list_faker.append(i)

    def get_phone_number(self) -> None:
        """
        Synthesize phone number columns in a pandas dataframe

        """

        phone_number = [
            i[0] for i in self.columns_with_assigned_entity if i[1] == "PHONE_NUMBER"
        ]

        for i in phone_number:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.phone_number()
                        if row == self.generation_mark
                        else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.phone_number() if not pd.isnull(row) else np.NaN
                    )
                )

                self.list_faker.append(i)

    def get_first_name(self) -> None:
        """
        Synthesize first name columns in a pandas dataframe

        """
        first_names = []
        first_name_person = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "PERSON"
            and (("first" in i[0].lower()) and ("name" in i[0].lower()))
        ]

        if "first_name_gender" in self.dataset.columns:
            for i in first_name_person:
                if self.generation_mark == "*":
                    for row in range(0, len(self.dataset[i])):
                        if self.dataset[i][row] == self.generation_mark:
                            if (
                                self.dataset["first_name_gender"][row] == "female"
                                or self.dataset["first_name_gender"][row]
                                == "mostly_female"
                            ):
                                self.dataset[i] = self.dataset[i].apply(
                                    lambda row: (self.faker.first_name_female())
                                )
                            elif (
                                self.dataset["first_name_gender"][row] == "male"
                                or self.dataset["first_name_gender"][row]
                                == "mostly_male"
                            ):
                                self.dataset[i] = self.dataset[i].apply(
                                    lambda row: (self.faker.first_name_male())
                                )
                            elif (
                                self.dataset["first_name_gender"][row] == "unknown"
                                or self.dataset["first_name_gender"][row] == "andy"
                            ):
                                self.dataset[i] = self.dataset[i].apply(
                                    lambda row: (self.faker.first_name())
                                )
                            else:
                                self.dataset[i][row]

                    self.list_faker.append(i)
                    return list(self.dataset[i])
                else:
                    for row in range(0, len(self.dataset[i])):
                        if not pd.isnull(self.dataset[i][row]):
                            if (
                                self.dataset["first_name_gender"][row] == "female"
                                or self.dataset["first_name_gender"][row]
                                == "mostly_female"
                            ):
                                self.dataset[i] = self.dataset[i].apply(
                                    lambda row: (self.faker.first_name_female())
                                )
                            elif (
                                self.dataset["first_name_gender"][row] == "male"
                                or self.dataset["first_name_gender"][row]
                                == "mostly_male"
                            ):
                                self.dataset[i] = self.dataset[i].apply(
                                    lambda row: (self.faker.first_name_male())
                                )
                            elif (
                                self.dataset["first_name_gender"][row] == "unknown"
                                or self.dataset["first_name_gender"][row] == "andy"
                            ):
                                self.dataset[i] = self.dataset[i].apply(
                                    lambda row: (self.faker.first_name())
                                )
                            else:
                                self.dataset[i][row] = np.NaN

                    self.list_faker.append(i)
                    return list(self.dataset[i])

            del self.dataset["first_name_gender"]

        else:
            for i in first_name_person:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.first_name() if not pd.isnull(row) else np.NaN
                    )
                )

                self.list_faker.append(i)
            return list(self.dataset[i])

    def get_last_name(self) -> None:
        """
        Synthesize last name columns in a pandas dataframe

        """

        last_name_person = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "PERSON"
            and (("last" in i[0].lower()) and ("name" in i[0].lower()))
        ]

        if len(last_name_person) > 0:
            for i in last_name_person:
                if self.generation_mark == "*":
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (self.faker.last_name() if row == "*" else row)
                    )
                    self.list_faker.append(i)

                else:
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.last_name() if not pd.isnull(row) else np.NaN
                        )
                    )

                    self.list_faker.append(i)

                return list(self.dataset[i])

        else:
            last_name_person = [
                i
                for i in self.dataset.columns
                if (("last" in i.lower()) and ("name" in i.lower()))
            ]
            for i in last_name_person:
                if self.generation_mark == "*":
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (self.faker.last_name() if row == "*" else row)
                    )
                    self.list_faker.append(i)
                else:
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.last_name() if not pd.isnull(row) else np.NaN
                        )
                    )

                    self.list_faker.append(i)

                return list(self.dataset[i])

    def get_email_address(self, names, last_names) -> None:
        """
        Synthesize email address columns in a pandas dataframe
        """

        email_address = [
            i[0] for i in self.columns_with_assigned_entity if i[1] == "EMAIL_ADDRESS"
        ]

        for i in email_address:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset.apply(
                    lambda row: (
                        names[row.name].lower()
                        + "."
                        + last_names[row.name].lower()
                        + "@"
                        + self.faker.free_email_domain()
                        if row == self.generation_mark
                        else row
                    ),
                    axis=1,
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset.apply(
                    lambda row: (
                        names[row.name].lower()
                        + "."
                        + last_names[row.name].lower()
                        + "@"
                        + self.faker.free_email_domain()
                        if not pd.isnull(row[i])
                        else row[i]
                    ),
                    axis=1,
                )

                self.list_faker.append(i)

    def get_city(self) -> None:
        """
        Synthesize city columns in a pandas dataframe

        """

        city = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "LOCATION"
            and (("city" in i[0].lower()) or ("cities" in i[0].lower()))
        ]

        for i in city:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.city() if row == self.generation_mark else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (self.faker.city() if not pd.isnull(row) else np.NaN)
                )

                self.list_faker.append(i)

    def get_state(self) -> None:
        """
        Synthesize state columns in a pandas dataframe

        """

        state = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "LOCATION" and ("state" in i[0].lower())
        ]

        for i in state:
            if len(self.dataset[i].iloc[0]) == 2:
                if self.generation_mark == "*":
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.state_abbr()
                            if row == self.generation_mark
                            else row
                        )
                    )
                    self.list_faker.append(i)
                else:
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.state_abbr() if not pd.isnull(row) else np.NaN
                        )
                    )

                    self.list_faker.append(i)
            else:
                if self.generation_mark == "*":
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.state() if row == self.generation_mark else row
                        )
                    )
                    self.list_faker.append(i)
                else:
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.state() if not pd.isnull(row) else np.NaN
                        )
                    )

                    self.list_faker.append(i)

    def get_url(self) -> None:
        """
        Synthesize url columns in a pandas dataframe

        """

        url = [i[0] for i in self.columns_with_assigned_entity if i[1] == "URL"]

        for i in url:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.url() if row == self.generation_mark else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (self.faker.url() if not pd.isnull(row) else np.NaN)
                )

                self.list_faker.append(i)

    def get_zipcode(self) -> None:
        """
        Synthesize zipcode columns in a pandas dataframe

        """

        zipcode = [i[0] for i in self.columns_with_assigned_entity if i[1] == "ZIPCODE"]

        if self.lang == "it":
            for i in zipcode:
                if self.generation_mark == "*":
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.postcode()
                            if row == self.generation_mark
                            else row
                        )
                    )
                    self.list_faker.append(i)
                else:
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.postcode() if not pd.isnull(row) else np.NaN
                        )
                    )

                    self.list_faker.append(i)
        else:
            for i in zipcode:
                if self.generation_mark == "*":
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.zipcode() if row == self.generation_mark else row
                        )
                    )
                    self.list_faker.append(i)
                else:
                    self.dataset[i] = self.dataset[i].apply(
                        lambda row: (
                            self.faker.zipcode() if not pd.isnull(row) else np.NaN
                        )
                    )

                    self.list_faker.append(i)

    def get_credit_card(self) -> None:
        """
        Synthesize credit card columns in a pandas dataframe

        """

        credit_card = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "CREDIT_CARD_NUMBER"
        ]

        for i in credit_card:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.credit_card_number()
                        if row == self.generation_mark
                        else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.credit_card_number()
                        if not pd.isnull(row)
                        else np.NaN
                    )
                )

                self.list_faker.append(i)

    def get_ssn(self) -> None:
        """
        Synthesize ssn columns in a pandas dataframe

        """

        ssn = [i[0] for i in self.columns_with_assigned_entity if i[1] == "US_SSN"]

        for i in ssn:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.ssn() if row == self.generation_mark else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (self.faker.ssn() if not pd.isnull(row) else np.NaN)
                )

                self.list_faker.append(i)

    def get_country(self) -> None:
        """
        Synthesize country columns in a pandas dataframe

        """

        country = [
            i[0]
            for i in self.columns_with_assigned_entity
            if i[1] == "LOCATION" and ("country" in i[0].lower())
        ]

        for i in country:
            if self.generation_mark == "*":
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (
                        self.faker.country() if row == self.generation_mark else row
                    )
                )
                self.list_faker.append(i)
            else:
                self.dataset[i] = self.dataset[i].apply(
                    lambda row: (self.faker.country() if not pd.isnull(row) else np.NaN)
                )

                self.list_faker.append(i)

    def get_columns_not_synthesized(self) -> None:
        """
        Get a list of all non-synthesized columns.

        """

        for i in self.columns_with_assigned_entity:
            if i[0] not in self.list_faker:
                self.columns_not_synthesized.append(i)

    def synthesis_message(self) -> None:
        """
        Get a message with synthesized and unsynthesized columns.

        """

        for col in self.list_faker:
            message = "Column " + green(col, "bold") + " synthesized with Faker."
            print(message)

        for col in self.columns_not_synthesized:
            message = "Column " + red(col[0], "bold") + " not synthesized with Faker."
            print(message)

    def get_faker_generation(self) -> None:
        """
        Get faker objects for columns in a pandas dataframe

        """
        self.get_columns_with_assigned_entity()
        self.get_address()
        self.get_phone_number()
        name = self.get_first_name()
        last_name = self.get_last_name()
        self.get_email_address(name, last_name)
        self.get_city()
        self.get_state()
        self.get_url()
        self.get_zipcode()
        self.get_credit_card()
        self.get_ssn()
        self.get_country()

        self.get_columns_not_synthesized()

        self.synthesis_message()



from typing import Any, Dict, List, Optional, Union

import gender_guesser.detector as gender
import numpy as np
import pandas as pd


from presidio_analyzer import (
    AnalyzerEngine,
    BatchAnalyzerEngine,
    PatternRecognizer,
    RecognizerRegistry,
)
from presidio_analyzer.nlp_engine import NlpEngineProvider
import spacy
from transformers import (
    AutoModelForTokenClassification,
    AutoTokenizer,
    logging,
    pipeline,
)

logging.set_verbosity_error()


# è importante eseguire prima assign_entities_with_presidio, poi
# assign_entities_manually, infine assisgn_entities_with_model


def split_name(df_input: Union[str, pd.DataFrame], name_of_column: str) -> pd.DataFrame:
    """
    Return a pandas dataframe where a given column with person's name is splitted into
    two columns: first_name and last_name

    Parameters
    ----------
    df_input : pd.DataFrame
        A pandas dataframe
    name_of_column : str
        name of the column which contains names that have to be splitted

    Returns
    -------
    pd.DataFrame
        A pandas dataframe with first_name and last_name columns
    """

    if not isinstance(df_input, pd.DataFrame):
        df_input = pd.read_csv(df_input)

    column = df_input[name_of_column]

    if column.isna().any():
        column.fillna("- -", inplace=True)

    list_name = []
    for i in column:
        list_name.append(i.split())

    for i in list_name:
        if len(i) < 2:
            i.append("-")

    name = []
    last_name = []
    for i in list_name:
        name.append(i[0])
        last_name.append(i[1])

    df_input["first_name"] = pd.Series(name)
    df_input["last_name"] = pd.Series(last_name)

    df_input = df_input.drop(name_of_column, axis=1)

    return df_input


def frequency(values: List, element: Any) -> float:
    """
    Calculate the frequency of an element in a list of values.

    Parameters
    ----------
    values : List
        List of values.
    element : Any
        Element to calculate the frequency of.

    Returns
    -------
    float
        Frequency of the element in the list.
    """
    return values.count(element) / len(values) if len(values) else 0


def en_add_address_entity(
    additional_addresses: Optional[List] = [],
) -> PatternRecognizer:
    """
    Return a customized presidio recognizer that can recognize ADDRESS entity.
    Some address-related words are already set, but user can add others.

    Parameters
    ----------
    additional_addresses : Optional[List], optional
        A list in which user can add new address-related words, by default []

    Returns
    -------
    PatternRecognizer
        A customized presidio recognizer
    """

    addresses = [
        "Street",
        "Rue",
        "Via",
        "Square",
        "Avenue",
        "Place",
        "Strada",
        "St",
        "Lane",
        "Road",
        "Boulevard",
        "Ln",
        "Rd",
        "HighwayDrive",
        "Av",
        "Hwy",
        "Blvd",
        "Corso",
        "Piazza",
        "Calle",
        "Plaza",
        "Avenida",
        "Rambla",
        "Vico",
        "C/",
    ]
    addresses = addresses + additional_addresses
    en_addresses_recognizer = PatternRecognizer(
        supported_language="en", supported_entity="ADDRESS", deny_list=addresses
    )

    return en_addresses_recognizer


def it_add_address_entity(
    additional_addresses: Optional[List] = [],
) -> PatternRecognizer:
    """
    Return a customized presidio recognizer that can recognize ADDRESS entity.
    Some address-related words are already set, but user can add others.

    Parameters
    ----------
    additional_addresses : Optional[List], optional
        A list in which user can add new address-related words, by default []

    Returns
    -------
    PatternRecognizer
        A customized presidio recognizer
    """

    addresses = [
        "Street",
        "Rue",
        "Via",
        "Square",
        "Avenue",
        "Place",
        "Strada",
        "St",
        "Lane",
        "Road",
        "Boulevard",
        "Ln",
        "Rd",
        "HighwayDrive",
        "Av",
        "Hwy",
        "Blvd",
        "Corso",
        "Piazza",
        "Calle",
        "Plaza",
        "Avenida",
        "Rambla",
        "Vico",
        "C/",
    ]
    addresses = addresses + additional_addresses
    it_addresses_recognizer = PatternRecognizer(
        supported_language="it", supported_entity="ADDRESS", deny_list=addresses
    )

    return it_addresses_recognizer


def get_gender(df_input: pd.DataFrame) -> pd.DataFrame:
    """Assign gender to each name in dataset

    Parameters
    ----------
    df_input : pd.DataFrame
        A pandas dataframe


    Returns
    -------
    pd.DataFrame
        A pandas dataframe with first_name_gender column
    """
    detector = gender.Detector(case_sensitive=False)
    first_name_gender = []

    for column in df_input.columns:
        if ("first" in column.lower()) and ("name" in column.lower()):
            for name in df_input[column]:
                if name is not np.NaN:
                    first_name_gender.append(detector.get_gender(name))
                else:
                    first_name_gender.append("Nan value")

    if len(first_name_gender) > 0:
        df_input["first_name_gender"] = pd.Series(first_name_gender)

    return df_input


class NamedEntityRecognizer:
    """
    A class used to recognize named entities in a dataset.

    Attributes
    ----------
    dataset : pd.DataFrame
        A pandas dataframe containing a sample of the dataset.
    object_columns : List
        A list of the object columns of the dataset.
    presidio_analyzer : BatchAnalyzerEngine
        A Presidio BatchAnalyzerEngine instance.
    assigned_entities_cols : List
        A list of the object columns of the dataset for which entities have been
         assigned.
    model : Any
        A pretrained nlp model downloaded from Hugging Face
    model_entities : Dict
        A dictionary whose keys are object column names and whose values are a list
        containing all the entities assigned to each value by the model
    dict_global_entities : Dict
        A dictionary whose keys have the same names of the dataframe columns and values
        are dictionaries in which the entity associated to the column and its confidence
        score are reported.
    en_spacy_model : Any
        An english spacy model
    it_spacy_model : Any
        An Italian spacy model


    Returns
    -------
    _type_
        _description_
    """

    original_dataset: pd.DataFrame
    dataset: pd.DataFrame
    object_columns: List
    presidio_analyzer: BatchAnalyzerEngine
    assigned_entities_cols: List
    model: Any
    model_entities: Dict
    dict_global_entities: Dict
    lang: str
    en_spacy_model: Any
    it_spacy_model: Any

    def __init__(
        self,
        df_input: Union[str, pd.DataFrame],
        data_sample: Optional[int] = 500,
        nan_filler: str = "?",
        lang: Optional[str] = "en",
        get_gender_option: Optional[bool] = False,
    ) -> "NamedEntityRecognizer":
        """
        Create a NamedEntityRecognizer instance.

        Parameters
        ----------
        df_input : Union[str, pd.DataFrame]
            A pandas dataframe or a path to a csv file.
        data_sample : Optional[int], optional
            Number of rows to sample from the dataframe, by default 500
        nan_filler : str, optional
            A string to fill the NaN values for object columns, by default "?"
        lang : str, optional
            Input language by default "en". Set this parameter to "it" to return better performance on Italian data.

        Returns
        -------
        NamedEntityRecognizer
            A NamedEntityRecognizer instance.
        """

        if not isinstance(df_input, pd.DataFrame):
            df_input = pd.read_csv(df_input)

        if get_gender_option == True:
            df_input = get_gender(df_input)

        self.dataset = df_input.sample(n=min(data_sample, df_input.shape[0]))
        self.object_columns = list(self.dataset.select_dtypes(["object"]).columns)
        # fill NaN values for object columns
        self.dataset.loc[:, self.object_columns] = self.dataset.loc[
            :, self.object_columns
        ].fillna(nan_filler)
        self.lang = lang

        self.presidio_analyzer = None
        self.model = None

        self.dict_global_entities = dict.fromkeys(list(self.dataset.columns))
        self.model_entities = {}
        self.assigned_entities_cols = []

        en_spacy_model_name = "en_core_web_lg"
        it_spacy_model_name = "it_core_news_lg"
        if self.lang == "en":
            if not spacy.util.is_package(en_spacy_model_name):
                spacy.cli.download(en_spacy_model_name)
            self.en_spacy_model = spacy.load(en_spacy_model_name)

        if self.lang == "it":
            if not spacy.util.is_package(it_spacy_model_name):
                spacy.cli.download(it_spacy_model_name)
            self.it_spacy_model = spacy.load(it_spacy_model_name)

    def set_presidio_analyzer(
        self,
        add_addresses_recognizer: Optional[bool] = True,
        additional_addresses: Optional[List] = [],
    ) -> None:
        """
        Set a Presidio BatchAnalyzer for the instance.

        Parameters
        ----------
        add_addresses_recognizer : Optional[bool], optional
            Whether to add a customized address recognizer, by default True
        additional_addresses : Optional[List], optional
            A list in which user can add new address-related words, by default []
        """

        if self.lang == "it":
            configuration = {
                "nlp_engine_name": "spacy",
                "models": [{"lang_code": "it", "model_name": "it_core_news_lg"}],
            }
            provider = NlpEngineProvider(nlp_configuration=configuration)
            nlp_engine_with_italian = provider.create_engine()

            analyzer = AnalyzerEngine(
                nlp_engine=nlp_engine_with_italian,
                supported_languages=["it"],
            )

            if add_addresses_recognizer:
                addresses_recognizer = it_add_address_entity(additional_addresses)
                analyzer.registry.add_recognizer(addresses_recognizer)

            self.presidio_analyzer = BatchAnalyzerEngine(analyzer_engine=analyzer)

        else:
            analyzer = AnalyzerEngine()

            if add_addresses_recognizer:
                addresses_recognizer = en_add_address_entity(additional_addresses)
                analyzer.registry.add_recognizer(addresses_recognizer)

            self.presidio_analyzer = BatchAnalyzerEngine(analyzer_engine=analyzer)

    def set_model(self) -> None:
        """
        Set a pretrained nlp model downloaded from Hugging Face
        (https://huggingface.co/dslim/bert-base-NER) used to recognize ORGANIZATION
        entities.

        Parameters
        ----------
        nlp_model : str, optional
            A NLP model name
        """
        if self.lang == "it":
            nlp_model = "osiria/bert-italian-uncased-ner"
        else:
            nlp_model = "dslim/bert-base-NER"

        tokenizer = AutoTokenizer.from_pretrained(nlp_model)
        model = AutoModelForTokenClassification.from_pretrained(nlp_model)
        self.model = pipeline("ner", model=model, tokenizer=tokenizer)

    def get_presidio_analyzer_results(self) -> List:
        """
        Get the results of the Presidio BatchAnalyzer: assign entities to each record
        in each columns of the dataset.

        Returns
        -------
        List
            A list containing the results of the analyzer.
        """
        if self.lang == "it":
            analyzer_results = list(
                self.presidio_analyzer.analyze_dict(
                    self.dataset.to_dict(orient="list"), language="it"
                )
            )
            return analyzer_results
        else:
            analyzer_results = list(
                self.presidio_analyzer.analyze_dict(
                    self.dataset.to_dict(orient="list"), language="en"
                )
            )
            return analyzer_results

    def assign_presidio_entities_list(self) -> None:
        """
        Get Presidio Analyzer results and assign entities to each object column of the
        dataset.
        """
        analyzer_results = self.get_presidio_analyzer_results()
        for col in analyzer_results:
            col_name = col.key
            if col_name in self.object_columns:
                # Get the list of entities for each record in the column
                entities_list = [
                    single_value_type[0].entity_type
                    for single_value_type in col.recognizer_results
                    if len(single_value_type) > 0
                ]
                # If the number of entities is more than 30% of the number of records,
                # assign the list to the column
                if len(entities_list) > 0.3 * self.dataset.shape[0]:
                    self.dict_global_entities[col_name] = entities_list
                    if col_name not in self.assigned_entities_cols:
                        self.assigned_entities_cols.append(col_name)

    def assign_location_entity(self) -> None:
        """
        Check whether the LOCATION entity is present among the entities assigned to the
        values in each column with assigned entities.
        If the LOCATION entity has been assigned to at least 10 percent of the values
        in that column,then the function assigns the LOCATION entity and the confidence
        score to that specific column
        """
        for col in self.assigned_entities_cols:
            entities_list = self.dict_global_entities[col]
            col_lower = col.lower()
            location_freq = frequency(entities_list, "LOCATION")
            if (
                ("LOCATION" in entities_list)
                and ("name" not in col_lower)
                and location_freq > 0.1
            ):
                self.dict_global_entities[col] = {
                    "entity": "LOCATION",
                    "confidence_score": location_freq,
                }

    def assign_entities_and_score(self) -> None:
        """
        Assign the most frequent entity and the confidence score to each object column
        with assigned entities.
        """
        for col in self.assigned_entities_cols:
            entities_list = self.dict_global_entities[col]
            if isinstance(entities_list, list):
                # Get the most frequent entity
                most_freq = max(set(entities_list), key=entities_list.count)
                self.dict_global_entities[col] = {
                    "entity": most_freq,
                    "confidence_score": frequency(entities_list, most_freq),
                }

    def assign_model_entities_list(self) -> None:
        """
        Assign entities to each object column which didn't get an entity from the
        Presidio Analyzer using the NLP model.
        """
        for col in self.object_columns:
            if self.dict_global_entities[col] is None:
                self.model_entities[col] = self.model(self.dataset[col].tolist())
                self.model_entities[col] = [
                    item["entity"]
                    for sublist in self.model_entities[col]
                    for item in sublist
                ]

    def assign_organization_entity(self) -> None:
        """
        Check whether the B-ORG entity is present among the entities assigned to the
        values in each column by the NLP model.

        If the B-ORG entity has been assigned to at least 10 percent of the values in
        that column, then the function assigns the LOCATION entity and the confidence
        score to that specific column.
        """
        for col in self.model_entities:
            entities_list = self.model_entities[col]
            organization_freq = frequency(entities_list, "B-ORG")
            if ("B-ORG" in entities_list) and organization_freq > 0.1:
                self.dict_global_entities[col] = {
                    "entity": "ORGANIZATION",
                    "confidence_score": organization_freq,
                }

    def assign_entities_manually(
        self, zipcode: Optional[bool] = True, credit_card: Optional[bool] = True
    ) -> None:
        """
        Assign ZIPCODE and CREDIT_CARD_NUMBER entities to each column of the dataset.

        Parameters
        ----------
        zipcode : Optional[bool], optional
            Whether to look for zipcodes in the column name, by default True
        credit_card : Optional[bool], optional
            Whether to look for credit card numbers in the column name, by default True
        """
        for col in self.dict_global_entities:
            col_lower = col.lower()
            if zipcode and (
                (("postal" in col_lower) and ("code" in col_lower))
                or ("zip" in col_lower)
                or ("cap" in col_lower)
            ):
                self.dict_global_entities[col] = {
                    "entity": "ZIPCODE",
                    "confidence_score": 1.0,
                }
            if credit_card and (
                (("credit" in col_lower) or ("card" in col_lower))
                and ("number" in col_lower)
                or (("carta" in col_lower) and ("credito" in col_lower))
            ):
                self.dict_global_entities[col] = {
                    "entity": "CREDIT_CARD_NUMBER",
                    "confidence_score": 1.0,
                }

    def assign_entities_with_presidio(self) -> None:
        """
        Set Presidio Analyzer and assign entities with a confidence score
        to each object column of the dataset.
        """
        self.set_presidio_analyzer()
        self.assign_presidio_entities_list()
        self.assign_location_entity()
        self.assign_entities_and_score()

    def assign_organization_entity_with_model(self) -> None:
        """
        Set NLP model and assign entities with a confidence score to each
        object column of the dataset.
        """
        self.set_model()
        self.assign_model_entities_list()
        self.assign_organization_entity()
